# 存在简单依赖第三方包的类和方法
import os
import re
from pprint import pprint
from copy import deepcopy
from typing import Dict, Union, Hashable, List, Optional, Callable, Literal
from tsc_base import dict_to_pair, pair_to_dict, merge_dict, get, any_value_dict, recur_opt_any
import xlwt
import openpyxl
from tqdm import tqdm
from openpyxl.worksheet.worksheet import Worksheet as openpyxl_Worksheet
from openpyxl.workbook.workbook import Workbook as openpyxl_Workbook
from xlwt.Workbook import Workbook as xlwt_Workbook
from xlwt.Worksheet import Worksheet as xlwt_Worksheet

NestedDict = Dict[Hashable, Union[int, 'NestedDict']]


class Excel:
    # 用于 openpyxl.utils.exceptions.IllegalCharacterError
    ILLEGAL_CHARACTERS_RE = re.compile(r'[\000-\010]|[\013-\014]|[\016-\037]')

    @staticmethod
    def get_excel_th(
        th_D: dict, 
        horizontal: bool = True, 
        start_r: int = 0, 
        start_c: int = 0, 
        rectangle: str = '$rectangle', 
        sort_key: Optional[dict] = None, 
        wild_card: str = '$*v', 
        max_deep: Optional[int] = None):
        """递归将dict转换为具有单元格位置坐标的excel表头, 用于 xlwt 等, 行列值从0开始

        Args:
            th_D (dict): 会把所有key拿出来作为表头, 然后value中增加 {rectangle:(开始行,结束行,开始列,结束列)}, 注意会覆盖原始value
            horizontal (bool, optional): 是否横向展开表头
            start_r (int, optional): 表头占据的开始行, 通常xlwt从0开始, openpyxl从1开始, 不过excel_add_sheet中进行了统一，所以这里都是从0开始
                行列是相对的，horizontal=False 这个相对于整个表而言就是列了。但是存到4元组后就是绝对值了
            start_c (int, optional): 表头占据的开始列
            rectangle (str, optional): 作为存储每个单元格的4角坐标的key, 也用于sort_key中的func, 不能与th_D中任意key一样
            sort_key (dict, optional): key:{rectangle:{'f':f(k,v),'r':bool},..}; 用于排序同一个字典中的k, 默认不排序
                f: 函数f输入key下面value中的k和v, 返回排序值, k会按照排序值排序, 例如 lambda t:t[0]
                r (bool, optional): 在 f 存在的情况下, 是否倒序排序, 不存在默认False
            wild_card (str, optional): 用于 sort_key 中匹配key的通配符, 如果在sort_key中找不到key则进入wild_card中的value, 不能与th_D中任意key一样
            max_deep (int, optional): 用于递归确定深度, 不能修改

        Returns:
            c, max_deep: int,int; 总占据列宽, 总占据行数
        """
        if max_deep is None:
            x = list(dict_to_pair(th_D))
            for i in x:
                assert len({rectangle, wild_card} & set(i[0])) == 0, 'th_D 中存在 rectangle/wild_card 标记, 请进行修改:' + str(i)
            max_deep = max([len(i[0]) for i in x])
        sort_key = {} if sort_key is None else sort_key
        if rectangle in sort_key:  # 排序
            par = sorted(th_D.items(), key=sort_key[rectangle]['f'], reverse=bool(sort_key[rectangle].get('r')))
        else:
            par = th_D.items()
        c = 0  # 深度优先得到上层的列宽
        for k, v in par:
            start_c_last = start_c + c  # 随着循环, 列的起始点在变化, 行的起始点不变
            r = max_deep
            if isinstance(v, dict):  # 随着递归, 行每次向下移动一行
                cc = Excel.get_excel_th(
                    v, horizontal, start_r+1, start_c_last, rectangle,
                    sort_key.get(k) or sort_key.get(wild_card), wild_card, max_deep-1)[0]
                if cc:
                    c += cc
                    r = 1  # 下面还要分之则只能占一行
                else:
                    c += 1  # 空字典
            else:
                c += 1  # 只占一列
                th_D[k] = v = {}
            v[rectangle] = (start_r, start_r+r-1, start_c_last, start_c+c-1)
            v[rectangle] = v[rectangle] if horizontal else (*v[rectangle][2:], *v[rectangle][:2])
        return c, max_deep

    @staticmethod
    def get_excel_table(
        doc_L: List[dict], 
        ignore_th_f: Callable = lambda t: t, 
        td_f: Callable = lambda t: t[1] if isinstance(t[1], (int, float)) else str(t[1]),
        ordered: bool = True,
        horizontal: bool = True, 
        rectangle: str = '$rectangle',
        **kw):
        """将多个dict转换为excel表格的单元格坐标, 用于生成excel表格

        Args:
            doc_L (list): [{..},..]; 被转换的dict列表
            ignore_th_f (func, optional): 输入 (key_L,value) 返回 (key_L,value) or None, None表示丢弃key_L这个表头
                用于修剪层次过深的表头, 这时候可能需要 td_f 优化dict在单元格中的展示形式
            td_f (func, optional): 用于优化单元格中的值, value是func, 输入 (key_L,value) 返回优化的展示 value
                小心处理每一种值格式, 因为 value 不能是 dict,list,tuple,set 等类型, 否则可能导致 excel 写入出错
            ordered (bool, optional): 是否保留 doc_L 中字段间的顺序(不是doc的顺序)，使用后速度会慢一些
            horizontal (bool, optional): 见 get_excel_th
            rectangle (str, optional): 见 get_excel_th
            **kw: 其他参数见 get_excel_th

        Returns:
            dict, list, list: th_D 的一个例子:
                {'inf_jcr': {'$rectangle': (0, 0, 1, 33),
                'Open Access': {'$rectangle': (1, 5, 29, 29)},
                '期刊分区': {'$rectangle': (1, 1, 1, 27),
                        'JCR分区': {'$rectangle': (2, 5, 27, 27)},}}
        """
        # 表头
        pair = []
        for p in dict_to_pair(merge_dict(doc_L, ordered=ordered)):
            p = ignore_th_f(p)
            if p is not None:
                pair.append(p)
        th_D = pair_to_dict(pair)
        Excel.get_excel_th(th_D, horizontal=horizontal, rectangle=rectangle, **kw)
        th_L = [(i[1], i[0][-2]) for i in dict_to_pair(th_D)]  # 与 td_L 格式一致
        # 表中值
        td_L = []  # [[((开始行,结束行,开始列,结束列),单元格值),..],..]; 与doc_L顺序一致
        for no, doc in enumerate(doc_L):
            coor_v_D = {}  # {(开始行,结束行,开始列,结束列):单元格值,..}
            for p in dict_to_pair(doc):
                p = ignore_th_f(p)  # (key_L,value); 用于获取行列位置, 位置靠前了值也会变
                # 检查这个值是否满足要求
                if p is None:
                    continue
                th = get(p[0], th_D)  # 对应的表头坐标
                if len(th) > 1:  # 没有到叶子结点, 说明doc存在大段的空
                    continue
                # 获取坐标
                _, r, _, c = th[rectangle]  # 右下角单元格坐标
                r, c = (r + no + 1, c) if horizontal else (r, c + no + 1)
                coor = (r, r, c, c)  # (开始行,结束行,开始列,结束列)
                # 保存坐标与单元格值
                if coor not in coor_v_D:  # 防止重复
                    coor_v_D[coor] = td_f((p[0], get(p[0], doc)))
            td_L.append(list(coor_v_D.items()))
        return th_D, th_L, td_L

    @staticmethod
    def excel_add_sheet(
        workbook: Union[openpyxl_Workbook, xlwt_Workbook],
        name: str, 
        th_L: list, 
        td_L: list, 
        package,
        index: Optional[int] = 0, 
        save_path: Optional[str] = None, 
        tqdm_f = None,
        auto_adjust_width: bool = False,
    ):
        """使用行列坐标 写入一页 excel 表格

        Args:
            workbook (obj): openpyxl.Workbook() or xlwt.Workbook(encoding='utf8')
            name (str): sheet 名称
            th_L (list): [[((开始行,结束行,开始列,结束列),单元格值),..],..]; 行列编号从0开始, 0行就是第一行
            td_L (list): [[((开始行,结束行,开始列,结束列),单元格值),..],..]; 行列编号从0开始, 0行就是第一行
            package : (xlwt>=1.3.0 import xlwt) or (openpyxl>=3.0.9 import openpyxl)
            index (int, optional): 只用于 openpyxl, 表示插入的 sheet 的位置, xlwt 只是追加
            save_path (str, optional): 写入文件的路径, 会自动添加后缀名
            tqdm_f : tqdm>=4.62.3 from tqdm import tqdm
            auto_adjust_width (bool, optional): 是否自动调整非行合并的单元格列宽以适应文本长度, 只有 openpyxl 支持

        Returns:
            workbook
        """
        if 'openpyxl' in str(type(workbook)):  # openpyxl - xlsx (打开效率更高)
            worksheet: openpyxl_Worksheet = workbook.create_sheet(name, index)
            # 写表头
            for coor, v in th_L:
                coor = [i+1 for i in coor]  # openpyxl 从1开始
                worksheet.merge_cells(start_row=coor[0], start_column=coor[2], end_row=coor[1], end_column=coor[3])
                cell = worksheet.cell(coor[0], coor[2])
                cell.value = Excel.ILLEGAL_CHARACTERS_RE.sub('', v) if type(v) == str else v
                cell.alignment = package.styles.Alignment(horizontal='center', vertical='center')
                cell.font = package.styles.Font(bold=True)
            # 单元格值
            par = tqdm_f(td_L, f'{name}-写入表格(openpyxl)') if tqdm_f else td_L
            for i in par:
                for coor, v in i:
                    cell = worksheet.cell(coor[0]+1, coor[2]+1)  # openpyxl 从1开始
                    cell.value = Excel.ILLEGAL_CHARACTERS_RE.sub('', v) if type(v) == str else v
                    cell.alignment = package.styles.Alignment(horizontal='center', vertical='center')
            if save_path:
                if not save_path.endswith('.xlsx'):
                    save_path = save_path + '.xlsx'
            else:
                save_path = None
            if auto_adjust_width:
                column_widths = {}
                for row in worksheet.iter_rows():  # 遍历所有行
                    for i, cell in enumerate(row):  # 遍历行中的所有单元格
                        if (  # 行的合并单元格跳过
                            isinstance(cell, package.cell.cell.MergedCell) or
                            i < len(row) - 1 and isinstance(row[i + 1], package.cell.cell.MergedCell)
                        ):
                            continue
                        if cell.value:
                            column_letter = cell.column_letter
                            current_width = column_widths.get(column_letter, 0)
                            new_width = len(str(cell.value))
                            column_widths[column_letter] = max(current_width, new_width)
                for column_letter, width in column_widths.items():  # 设置列宽
                    worksheet.column_dimensions[column_letter].width = width + 2
        else:  # xlwt - xls (写入快6,7倍)
            worksheet: xlwt_Worksheet = workbook.add_sheet(name)
            # 单元格值
            style = package.XFStyle()
            alignment = package.Alignment()
            alignment.horz = package.Alignment.HORZ_CENTER
            alignment.vert = package.Alignment.VERT_CENTER
            style.alignment = alignment
            par = tqdm_f(td_L, f'{name}-写入表格(xlwt)') if tqdm_f else td_L
            for i in par:
                for coor, v in i:
                    worksheet.write_merge(*coor, v, style)
            # 写表头
            font = package.Font()
            font.bold = True
            style.font = font
            for coor, v in th_L:
                worksheet.write_merge(*coor, v, style)
            if save_path:
                if not save_path.endswith('.xls'):
                    save_path = save_path + '.xls'
            else:
                save_path = None
        if save_path:
            if os.path.dirname(save_path) and not os.path.exists(os.path.dirname(save_path)):
                os.makedirs(os.path.dirname(save_path))
            workbook.save(save_path)
        return worksheet

    @classmethod
    def get_excel_table_with_side(cls, doc_L: List[dict], side_th: NestedDict = None, **kwargs) -> dict:
        """将多个dict转换为excel表格的单元格坐标, 用于生成excel表格, 侧边栏表头
        最终用于 excel_add_sheet 的方式：
        th_L = ret['th_L'] + ret['side_th_L']
        td_L = ret['td_L']
        注意：主表头的最后一行一定不是合并单元格，侧边栏表头的最后一行是不是都行

        Args:
            doc_L (List[dict]): [{..},..]; 被转换的dict列表
            side_th (NestedDict, optional): 侧边栏表头, 会在表格左边或上边生成一个表格, 用于存储一些额外信息
            **kwargs: 其他参数见 get_excel_table

        Returns:
            dict: 参考 get_excel_table
        """
        if side_th:
            doc_num = any_value_dict(side_th, lambda v1, v2: v1 + v2, 0, True)  # 有非int会报错
            assert doc_num > 0, 'side_th 中没有文档!'
            assert doc_num == len(doc_L), f'从 side_th 得到的文档数 {doc_num} 与 doc_L 中的 {len(doc_L)} 不一致'
            side_th_high = max(len(i[0]) for i in dict_to_pair(side_th))  # 侧边栏表头的最大深度
            side_th_ext = deepcopy(side_th)
            recur_opt_any(side_th_ext, lambda k, t, v: ({i: i for i in range(v)}, 0, 1) if isinstance(v, int) else (v, 0, 0))
            # 表头参数
            kwargs['ordered'] = True  # doc_L 顺序不能变
            side_kw = kwargs.copy()
            side_kw['horizontal'] = not kwargs.get('horizontal', True)  # 侧边栏表头的横竖是相反的
            kwargs['start_c'] = kwargs.get('start_c', 0) + side_th_high  # 因为 start_c 是相对的，永远是列
            # 核心表格计算
            th_D, th_L, td_L = cls.get_excel_table(doc_L, **kwargs)
            th_high = max(len(i[0]) for i in dict_to_pair(th_D)) - 1  # 核心表格的最大深度
            # 侧边栏表格计算
            side_kw['start_c'] = side_kw.get('start_c', 0) + th_high
            side_th_D, side_th_L, side_td_L = cls.get_excel_table([side_th_ext], **side_kw)
            
            # 去除最后一行，由side_th_ext引入的辅助行
            # 删除叶子
            recur_opt_any(side_th_D, lambda k, t, v: (v, 1, 0) if isinstance(v, dict) and len(v) <= 1 else (v, 0, 0))
            # 确定补全行或列的位置，防止合并单元格不全
            if side_kw['horizontal']:
                complete_no = 1  # 补全行
            else:
                complete_no = 3  # 补全列
            max_no = side_kw.get('start_r', 0) + side_th_high - 1
            # 补全行或列
            rectangle = kwargs.get('rectangle', '$rectangle')
            
            def opt(keys, types, value):
                if isinstance(value, dict) and len(value) == 1:
                    coor = value[rectangle]
                    value[rectangle] = coor[:complete_no] + (max_no,) + coor[complete_no+1:]
                    return value, 0, 1
                return value, 0, 0
            recur_opt_any(side_th_D, opt)
            # 重新生成侧边栏表头坐标
            side_th_L = [(i[1], i[0][-2]) for i in dict_to_pair(side_th_D)]
        else:
            th_D, th_L, td_L = cls.get_excel_table(doc_L, **kwargs)
            side_th_D, side_th_L, side_td_L = {}, [], []
        return {
            'th_D': th_D,
            'th_L': th_L,
            'td_L': td_L,
            'side_th_D': side_th_D,
            'side_th_L': side_th_L,
        }


if __name__ == '__main__':
    # Excel
    print('=' * 10, 'Excel')
    doc_L = [{
        'a111111你': {'b11': i, 'b2': 2, 'b3': {'c1': 1, 'c22222': False}},
        'a2': '123123123',
        'a3': {},
        'a4': None,
        'a5': [],
    } for i in range(10)]
    openpyxl_wb = openpyxl.Workbook()
    xlwt_wb = xlwt.Workbook(encoding='utf8')
    ignore_th_f = lambda t: None if t[1] == {} or t[1] == [] else t
    
    for horizontal, name in [(True, '横向'), (False, '纵向')]:
        th_D, th_L, td_L = Excel.get_excel_table(doc_L, ignore_th_f=ignore_th_f, horizontal=horizontal)
        print('th_D:')
        pprint(th_D)
        Excel.excel_add_sheet(openpyxl_wb, f'{name}', th_L, td_L, openpyxl, 0, 'test/openpyxl', tqdm, auto_adjust_width=True)
        Excel.excel_add_sheet(xlwt_wb, f'{name}', th_L, td_L, xlwt, 0, 'test/xlwt', tqdm, auto_adjust_width=True)
        
        print('=' * 5, 'add side:')
        ret = Excel.get_excel_table_with_side(doc_L, side_th={
            'test': {'a': 1, 'b'*10: 2, 'c': {'d': 1, 'e': len(doc_L) - 4}},
        }, ignore_th_f=ignore_th_f, horizontal=horizontal)
        print('side_th_D:')
        pprint(ret['side_th_D'])
        th_L = ret['th_L'] + ret['side_th_L']
        td_L = ret['td_L']
        Excel.excel_add_sheet(openpyxl_wb, f'{name}2', th_L, td_L, openpyxl, 0, 'test/openpyxl', tqdm, auto_adjust_width=True)
        Excel.excel_add_sheet(xlwt_wb, f'{name}2', th_L, td_L, xlwt, 0, 'test/xlwt', tqdm, auto_adjust_width=True)
        
        print()



import os
import datetime
import operator
from functools import reduce
import ast


class ParaTex:
    def __init__(self, leftParas, upParas, blacklist=None, whitelist=None, ignoreP=None):
        r"""
        参数, 参数名, 参数值, 参数值名 都会被转为字符串类型(为了readTableE), 它们都不可为空和None, 不能包括\符号, 以及{}#$%^&_ 防止latex特殊符号出错.
        考虑到转化为字符串, 因此数值型的str参数需要书写标准, 否则可能与数值型参数不相等.
        @param leftParas: [[参数,参数名,(参数值,..),(参数值名,..)],..], 表格从左至右的参数
        @param upParas: [[参数,参数名,(参数值,..),(参数值名,..)],..], 表格从上至下的参数
        @param blacklist: [{参数:(参数值,..),..},..], 冲突时优先级高于白名单, nextParas 会跳过黑名单输出
        @param whitelist: [{参数:(参数值,..),..},..], 只包含空 {} 会使所有参数为非计算
        @param ignoreP: [参数,..], nextParas 不输出的参数, 用于一次计算返回多个值, setParaV_c的内容值c要按照这个参数及其参数值顺序全排列
        """
        assert len(leftParas[0][2]) == len(leftParas[0][3]) > 0, 'leftParas 格式错误(要求%d==%d>0)!' % (
            len(leftParas[0][2]), len(leftParas[0][3]))
        assert len(upParas[0][2]) == len(upParas[0][3]) > 0, 'upParas 格式错误(要求%d==%d>0)!' % (
            len(upParas[0][2]), len(upParas[0][3]))
        # 转化为str
        haveEmpty = [False]
        leftParas = self._allToString(leftParas, haveEmpty)
        assert not haveEmpty[0], 'leftParas 存在空字符串!'
        upParas = self._allToString(upParas, haveEmpty)
        assert not haveEmpty[0], 'upParas 存在空字符串!'
        if blacklist:
            blacklist = self._allToString(blacklist, haveEmpty)
            assert not haveEmpty[0], 'blacklist 存在空字符串!'
        if whitelist:
            whitelist = self._allToString(whitelist, haveEmpty)
            assert not haveEmpty[0], 'whitelist 存在空字符串!'
        if ignoreP:
            ignoreP = self._allToString(ignoreP, haveEmpty)
            assert not haveEmpty[0], 'ignoreP 存在空字符串!'
        # 获得全排列参数
        paras_v = []
        for i in leftParas:
            paras_v.append([i[0], i[2]])
        for i in upParas:
            paras_v.append([i[0], i[2]])
        all_p = self._all_permutation(paras_v)  # 全排列参数
        all_p_v8m = {i: ['', False if whitelist else True] for i in all_p}

        self._leftParas = leftParas
        self._upParas = upParas
        self._paras_v = paras_v  # [[参数,(参数值,..)],..]
        self._paras_v_D = dict(paras_v)  # 用于 setParaV_c, readTableE
        self._all_p = all_p  # [(排列参数),..]
        self._all_p_v8m = all_p_v8m  # {(排列参数):[值,是否计算],..}
        self._index = -1
        self._alreadyP = set()  # 已经 next 过的排列参数, 用于辅助 ignoreP
        self._ignoreP = ignoreP if ignoreP else []  # 用于 setParaV_c 按序输入
        self._ignoreP_s = set(self._ignoreP)  # 用于 nextParas 判断是否去除
        assert len(self._ignoreP_s & set([i[0] for i in paras_v])) == len(self._ignoreP_s), 'ignoreP存在参数不在表中!'

        # 标注白名单
        if whitelist:
            for l_s in whitelist:
                self._list_filter(l_s, True)
        # 标注黑名单
        if blacklist:
            for l_s in blacklist:
                self._list_filter(l_s, False)

    def _allToString(self, x, haveEmpty: list = None, sfs=True):
        '''
        将非 tuple, list, set, dict 类型递归转为 str 类型, 除了 None.
        @param x: tuple|list|set|dict|str|int|float|bool|None
        @param haveEmpty: list, haveEmpty[0] 表示是否包含空字符串
        @param sfs: bool, 是否将str转float再转str, 可以部分防止因为str中数字多写.或0之类导致不相等, 但是如果故意想不相等反而导致反效果
        @return: tuple|list|set|dict|str|int|float|bool|None
        '''
        if isinstance(x, tuple) or isinstance(x, list) or isinstance(x, set):
            y = []
            for i in x:
                y.append(self._allToString(i, haveEmpty))
            if isinstance(x, tuple):
                y = tuple(y)
            if isinstance(x, set):
                y = set(y)
            return y
        elif isinstance(x, dict):
            y = {}
            for k, v in x.items():
                y[self._allToString(k, haveEmpty)] = self._allToString(v, haveEmpty)
            return y
        elif x is None:
            return None
        else:
            if sfs and isinstance(x, str):
                try:
                    if '.' in x:
                        x = float(x)
                    else:
                        x = int(x)
                except:
                    ...
            x = str(x)
            if not x and haveEmpty:
                haveEmpty[0] = True
            return x

    def _list_filter(self, l_s, white=True):
        """
        黑白名单过滤器.
        @param l_s: {参数:(参数值,..),..}
        @param white: 是不是白名单
        """
        if not l_s:
            return
        l_all = []  # [[参数,(参数值,..)],..], 加入没有的参数
        for i in self._paras_v:
            if i[0] in l_s:
                v = l_s[i[0]]
                assert isinstance(v, tuple), '参数值元组类型不是tuple!'
                assert len(set(v) & set(i[1])) == len(v), '名单参数值不在范围内!' + str(v) + ' not in ' + str(i[1])
                l_all.append([i[0], v])
            else:
                l_all.append(i)
        l_all_p = self._all_permutation(l_all)
        for i in l_all_p:
            self._all_p_v8m[i][1] = white

    @staticmethod
    def _all_permutation(paras, all=None, seq=None):
        """
        全排列算法.
        @param paras: [[参数,(参数值,..)],..]
        @return: [(排列参数),..]
        """
        if not paras:
            return []
        if seq is None:
            seq = []
        if all is None:
            all = []
        para_v_c = paras[0]
        for v in para_v_c[1]:
            if len(paras) <= 1:
                all.append(tuple(seq + [v]))
            else:
                ParaTex._all_permutation(paras[1:], all, seq + [v])
        return all

    def nextParas(self):
        """
        返回下一组要计算的参数.
        @return: {参数:参数值,..}, 若有ignoreP则返回部分参数
        """
        while True:
            self._index += 1
            if len(self._all_p) <= self._index:
                self._index = -1  # 重置
                self._alreadyP = set()
                return
            seq = self._all_p[self._index]
            seq_i = tuple([seq[i] for i, x in enumerate(self._paras_v) if x[0] not in self._ignoreP_s])
            if seq_i in self._alreadyP:
                continue
            if self._all_p_v8m[seq][1]:
                self._alreadyP.add(seq_i)
                paras_v = {x[0]: seq[i] for i, x in enumerate(self._paras_v) if x[0] not in self._ignoreP_s}
                yield paras_v

    def setParaV_c(self, paras_v: dict, c, setNoCompute=False):
        """
        根据一组参数设置对应的表内容, 内容符号需要注意, 用于latex.
        @param paras_v: {参数:参数值,..}, 允许多于表格参数
        @param c: other|list, 内容, 如果是列表说明有多个参数一起赋值, 按ignoreP参数全排列
        @param setNoCompute: bool, 是否设置为非计算, 即nextParas是否输出
        @return: dict, 插入成功的参数组数量, 插入重复的数量, 错误的数量,... 3者没有交集
        """
        seq = []  # 参数序列
        assert paras_v, 'paras_v 参数不能为空!'
        if isinstance(c, list) and len(c) == 1:
            c = c[0]
        out = {
            'success': 0,  # 成功不重复写入的参数组数量
            'repeat': 0,  # 和已有值重复的写入数量
            'error': 0,  # 无法写入的参数组数量
            'error_seq': [],  # 无法写入的参数组
        }
        paras_v = self._allToString(paras_v)
        c = self._allToString(c)
        if isinstance(c, list):
            ignoreP_num = reduce(operator.mul, [len(i[1]) for i in self._paras_v if i[0] not in paras_v], 1)  # 最小是1
            assert ignoreP_num == len(c), '忽律参数值数量与内容不匹配(%d!=%d)!' % (ignoreP_num, len(c))
            ignoreP = [(i, self._paras_v_D[i]) for i in self._ignoreP]  # [[参数,(参数值,..)],..], 没填的参数
            ignoreP_per = self._all_permutation(ignoreP)  # [(排列参数),..]
            for k, seq in enumerate(ignoreP_per):
                p_v = {ignoreP[i][0]: seq[i] for i in range(len(ignoreP))}
                p_v.update(paras_v)
                for i, j in self.setParaV_c(p_v, c[k], setNoCompute).items():
                    out[i] += j
        else:
            assert len(set(paras_v) & set(self._paras_v_D)) == len(self._paras_v), '参数数量不正确(%d!=%d)!' % (
                len(set(paras_v) | set(self._paras_v_D)), len(self._paras_v))
            for para, _ in self._paras_v:
                seq.append(paras_v[para])
            seq = tuple(seq)
            if seq in self._all_p_v8m:
                if setNoCompute:
                    self._all_p_v8m[seq][1] = False
                if self._all_p_v8m[seq][0] == c:
                    out['repeat'] += 1
                else:
                    self._all_p_v8m[seq][0] = c
                    out['success'] += 1
            else:
                out['error'] += 1
                out['error_seq'].append(seq)
        return out

    def getParasG_n(self, all=False):
        """
        获取一共有多少组参数.
        @param all: bool, 是不是全排列数量, 如果为False直接返回nextParas数量, 否则是全排列数量
        @return: int, 参数全排列有多少
        """
        allN = len(self._all_p)
        if not all:
            allN = 0
            for i in self.nextParas():
                allN += 1
        return allN

    def getParaVaddN(self, paraV: dict):
        """
        为参数和参数值获取参数名和参数值名, 若同一参数相同参数值一样则取第一个.
        @param paraV: {参数:参数值,..}
        @return: [(参数,参数名,参数值,参数值名),..]
        """
        out = []
        paraV = self._allToString(paraV)
        for i in self._leftParas + self._upParas:
            if i[0] not in paraV:
                continue
            k = j = None
            for k, j in enumerate(i[2]):
                if j == paraV[i[0]]:
                    break
            out.append((i[0], i[1], j, i[3][k]))
        return out

    def getTable(self, caption='Test', left_bf=False, up_bf=True, if_bf=max, dec=None, useV=False, path=None,
                 spMark='=' * 50, pathRepCheck=True):
        r"""
        输出表格, 需要latex包:
        \usepackage{multirow}  % 如果用到合并单元格
        \usepackage{CJKutf8}  % 如果包含中文
        @param caption: str, 标题
        @param left_bf: bool, 是否每行特殊值加粗
        @param up_bf: bool, 是否每列特殊值加粗
        @param if_bf: func, 是否加粗比较, 输入列表, 返回应该加粗的一个值. 一般是 max 或 min
        @param dec: None|int, 统一保留几位小数. 在加粗判断之后执行, 可能因为四舍五入导致看起来一样的值没有加粗
        @param useV: 是否使用参数和参数值作为表头(而不是参数名和参数值名)
        @param path: str, 保存路径, 追加模式写入
        @param spMark: str, 保存表格前输出的分割符号, 作为一行
        @param pathRepCheck: bool, 是否检查日志中已有一样的表格, 有则不输出, 这需要读取匹配全日志
        @return: str, latex表格
        """
        left_num = len(self._leftParas)  # 行参数数
        up_num = len(self._upParas)  # 列参数数
        left_numAll = reduce(operator.mul, [len(i[2]) for i in self._leftParas], 1)  # 总行数, 不含表头
        up_numAll = reduce(operator.mul, [len(i[2]) for i in self._upParas], 1)  # 总列数, 不含表头
        if len(self._upParas) == 1:
            up_numAll_1 = up_numAll
        else:
            up_numAll_1 = up_numAll // len(self._upParas[-1][2])  # 除最后一层 总列数
        cc = ('|' + 'c' * len(self._upParas[-1][2])) if len(self._upParas) > 1 else 'c'  # 合并列的表头
        assert left_numAll * up_numAll == len(self._all_p), '行列相乘不等于总数!'
        if useV:
            header1 = 0
            header2 = 2
        else:
            header1 = 1
            header2 = 3
        # 标题替换特殊字符
        caption = caption.replace('\\', r' ')
        caption = caption.replace('#', r'\#').replace('$', r'\$')
        caption = caption.replace('%', r'\%').replace('^', r'\^')
        caption = caption.replace('&', r'\&').replace('_', r'\_')
        caption = caption.replace('{', r'\{').replace('}', r'\}')
        # 头部
        table = '\\begin{table}[h]\n\\centering\n\\caption{%s}\n\\begin{tabular}{|%s} \\hline\n' % (
            caption, 'c|' * left_num + (cc * up_numAll_1).strip('|') + '|')
        # 表行头
        l = 1
        for i in range(up_num - 1):
            table += '\t'
            if left_num > 1:  # 需要占多行
                table += r'\multicolumn{%d}{|r|}{\bf{%s}}' % (left_num, str(self._upParas[i][header1]))
            else:  # 只有1列不需要占多行
                table += r'\bf{%s}' % str(self._upParas[i][header1])
            for k in range(l):  # 参数值名
                for j in self._upParas[i][header2]:
                    ll = up_numAll // len(self._upParas[i][3]) // l
                    if ll > 1:
                        x = r' & \multicolumn{%d}{|c|}{%s}' % (ll, str(j))
                    else:
                        x = ' & ' + str(j)
                    table += x
            if left_num > 1:
                table += r'\\ \cline{%d-%d}' % (left_num, left_num + up_numAll)
            else:
                table += r'\\ \hline'
            table += '\n'
            l *= len(self._upParas[i][3])
        # 表行头 最后一行
        table += '\t\\bf{%s}' % str(self._leftParas[0][header1])
        for i in range(1, left_num):
            if i == left_num - 1:
                table += r' & $_\textbf{%s}$' % str(self._leftParas[i][header1])
            else:
                table += r' & \bf{%s}' % str(self._leftParas[i][header1])
        table += r'\verb|\|$^\textbf{%s}$' % str(self._upParas[-1][header1])
        for i in range(l):
            for j in self._upParas[-1][header2]:
                table += ' & ' + str(j)
        table += '\\\\ \\hline \n'
        # 加粗
        all_v8bf_L = [[self._all_p_v8m[i][0], False] for i in self._all_p]  # [[值,是否加粗],..], 按 self._all_p 顺序

        def bf(y):  # 加粗函数, 返回对应索引
            try:
                x = [float('0' + str(i)) for i in y]
            except:
                x = [str(i) for i in y]
            xx = if_bf(x)
            return (lambda x: [j for j, i in enumerate(x) if i == xx])(x)

        if left_bf:
            for i in range(left_numAll):
                v_L = [i[0] for i in all_v8bf_L[i * up_numAll: (i + 1) * up_numAll]]
                for j in bf(v_L):
                    all_v8bf_L[j + i * up_numAll][1] = True
        if up_bf:
            for i in range(up_numAll):
                v_L = []
                x_L = []
                for j in range(left_numAll):
                    x = j * up_numAll + i
                    v_L.append(all_v8bf_L[x][0])
                    x_L.append(x)
                i_L = bf(v_L)
                for j in i_L:
                    all_v8bf_L[x_L[j]][1] = True
        # 内容
        cs_L = [''] * left_numAll  # 每一行值, 行表头s-内容-尾部

        def bf_v(i, j):  # 根据行列和加粗情况返回文本值
            v, b = all_v8bf_L[i * up_numAll + j]
            t = str(v).strip()
            if dec:  # 小数
                try:
                    t = ('%.' + str(dec) + 'f') % float(t)
                except:
                    ...
            if b and t:
                t = r'\bf{%s}' % t
            return t

        for i in range(left_numAll):  # 内容值
            cs_L[i] += bf_v(i, 0)
            for j in range(1, up_numAll):
                cs_L[i] += ' & ' + bf_v(i, j)
        default_s = '& '
        cs_L = [[default_s] * left_num + [i, r'\\'] for i in cs_L]  # 扩展多列
        l = left_numAll
        for j in range(left_num):  # 开头值
            l = l // len(self._leftParas[j][2])
            for i in range(left_numAll):
                if i % l == 0:  # l 步输出一个
                    ll = i // l % len(self._leftParas[j][2])  # 输出第几个值
                    if l > 1:
                        cs_L[i][j] = r'\multirow{%d}{*}{%s} ' % (l, self._leftParas[j][header2][ll]) + str(cs_L[i][j])
                    else:
                        cs_L[i][j] = r'%s ' % (str(self._leftParas[j][header2][ll])) + str(cs_L[i][j])
                    if i < left_numAll - 1 and j > 0:
                        for k in range(0, j):
                            if cs_L[i + 1][k] != default_s:
                                cs_L[i][-1] += r' \cline{%d-%d}' % (k + 1, left_num + up_numAll)
                                break
        cs_L[-1][-1] += r' \hline'
        for i in range(left_numAll):
            table += '\t' + ''.join(cs_L[i]) + '\n'
        # 尾部
        table += '\\end{tabular}\n\\end{table}'
        # 写入
        write = True
        if pathRepCheck and path and os.path.exists(path):
            with open(path, 'r', encoding='utf-8') as r:
                if table in r.read():
                    write = False
        if path and write:
            with open(path, 'a', encoding='utf-8') as w:
                w.write(spMark + '\n')
                w.write('time: ' + str(datetime.datetime.now()) + '\n')
                w.write('type: latex table\n')
                w.write(table + '\n')
        return table

    def getTableE(self, caption='Test', dec=None, useV=True, path=None, spMark='=' * 50, pathRepCheck=True):
        r"""
        输出 excel 表格, \t分割, 无加粗, 无合并单元格.
        @param caption: str, 标题
        @param dec: None|int, 统一保留几位小数
        @param useV: 是否使用参数和参数值作为表头(而不是参数名和参数值名), 只有 True 才能使用 readTableE
        @param path: str, 保存路径, 追加模式写入
        @param spMark: str, 保存表格前输出的分割符号, 作为一行
        @param pathRepCheck: bool, 是否检查日志中已有一样的表格, 有则不输出, 这需要读取匹配全日志
        @return: str, excel 表格
        """
        left_num = len(self._leftParas)  # 行参数数
        up_num = len(self._upParas)  # 列参数数
        left_numAll = reduce(operator.mul, [len(i[2]) for i in self._leftParas], 1)  # 总行数, 不含表头
        up_numAll = reduce(operator.mul, [len(i[2]) for i in self._upParas], 1)  # 总列数, 不含表头
        assert left_numAll * up_numAll == len(self._all_p), '行列相乘不等于总数!'
        if useV:
            header1 = 0
            header2 = 2
        else:
            header1 = 1
            header2 = 3
        # 头部
        table = caption + '\n'
        # 表行头
        l = 1
        for i in range(up_num - 1):
            table += '\t' * (left_num - 1)
            table += '%s' % str(self._upParas[i][header1])
            for k in range(l):  # 输出参数值名
                for j in self._upParas[i][header2]:
                    ll = up_numAll // len(self._upParas[i][3]) // l
                    if ll > 1:
                        x = '\t%s%s' % (str(j), '\t' * (ll - 1))
                    else:
                        x = '\t' + str(j)
                    table += x
            table += '\n'
            l *= len(self._upParas[i][3])
        # 表行头 最后一行
        table += '%s' % str(self._leftParas[0][header1])
        for i in range(1, left_num):
            table += '\t%s' % str(self._leftParas[i][header1])
        table += r'\%s' % str(self._upParas[-1][header1])
        for i in range(l):
            for j in self._upParas[-1][header2]:
                table += '\t' + str(j)
        # 内容
        cs_L = [''] * left_numAll  # 每一行, 行表头s-内容-尾部
        all_v_L = [self._all_p_v8m[i][0] for i in self._all_p]  # [值,..], 按 self._all_p 顺序

        def bf_v(i, j):  # 根据小数位数返回文本值
            v = all_v_L[i * up_numAll + j]
            t = str(v).strip()
            if dec:  # 小数
                try:
                    t = ('%.' + str(dec) + 'f') % float(t)
                except:
                    ...
            return t

        for i in range(left_numAll):  # 内容值
            cs_L[i] += bf_v(i, 0)
            for j in range(1, up_numAll):
                cs_L[i] += '\t' + bf_v(i, j)
        default_s = '\t'
        cs_L = [[default_s] * left_num + [i] for i in cs_L]  # 扩展多列
        l = left_numAll
        for j in range(left_num):  # 开头值
            l = l // len(self._leftParas[j][2])
            for i in range(left_numAll):
                if i % l == 0:  # l 步输出一个
                    ll = i // l % len(self._leftParas[j][2])  # 输出第几个值
                    cs_L[i][j] = r'%s ' % (str(self._leftParas[j][header2][ll])) + str(cs_L[i][j])
        for i in range(left_numAll):
            table += '\n' + ''.join(cs_L[i])
        # 写入
        write = True
        if pathRepCheck and path and os.path.exists(path):
            with open(path, 'r', encoding='utf-8') as r:
                if table in r.read():
                    write = False
        if path and write:
            with open(path, 'a', encoding='utf-8') as w:
                w.write(spMark + '\n')
                w.write('time: ' + str(datetime.datetime.now()) + '\n')
                w.write('type: excel table\n')
                w.write(table + '\n')
        return table

    def readTableE(self, path, setNoCompute=False, spMark='=' * 50, typeMark='type: excel table', lastOne=False,
                   simplifyOut=True):
        r'''
        用于将日志中 getTableE 输出的表格中的内容读取到表格中, 如果内容为空字符或None则不读取, 如果表格不规范可能导致读取程序报错或作为错误表格.
        读取的表格格式必须与 getTableE 输出到文件中的一致, 参数/参数名/参数值/参数值名都不可为空字符串和None, 不能包括\符号.
        getTableE 输出的必须是参数(值)(useV=True), 而不是参数值名(useV=False), 否则读取失败.
        末尾制表符不能少.
        @param path: str, 日志路径
        @param setNoCompute: bool, 日志中的值是否设置为非计算, 即nextParas是否输出
        @param spMark: str, 每组参数的上分隔符, 第一个分隔符上面的不看
        @param typeMark: str, 从 typeMark 开始下面第一行出现\t的地方开始读取表格, 到无\t行结束
        @param lastOne: bool, 是否只用日志中最后一个表格, False 表示日志中所有表格从上到下覆盖叠加, 如果为True前面的表格有效也不会检测
        @param simplifyOut: bool, 用于简化返回参数, 如果为True则不返回字符数超过100的统计参数
        @return: dict, 统计, 如果有的话 success+repeat = 不计算的参数组数量
        '''
        out = {
            'success': 0,  # 成功不重复写入的参数组数量
            'repeat': 0,  # 和已有值重复的写入数量
            'error_table': 0,  # 无效表格的数量
            'mark': 0,  # typeMark 的数量, 包含无效表格
        }
        try:
            with open(path, 'r', encoding='utf-8') as r:
                logs = r.read().split(spMark)[1:]
        except:
            return out
        tables = []  # [[[第一行第一格，..],..],..], 多个表格
        for k, log in enumerate(logs):  # 读取所有mask为表格的
            if typeMark not in log:
                continue
            out['mark'] += 1
            log = log.split(typeMark)[1].strip('\r\n').split('\n')
            start = False
            for line in log:
                if not start and '\t' in line:
                    start = True
                    tables.append([])  # [[第一行第一格，..],..]
                if start:
                    if '\t' in line:
                        tables[-1].append([i.strip() for i in line.split('\t')])  # [第一行第一格，..]
                    else:
                        break
        if lastOne:  # 只取最后一个
            tables = [tables[-1]]
        for ii, table in enumerate(tables):
            up_paras = []  # [参数,..], 上面参数(名)
            left_paras = []  # [参数,..], 左边参数(名)
            up_paras_V = []  # [[参数值,..],..], 上面参数值(名)
            left_paras_V = []  # [[参数值,..],..], 左边参数值(名)
            up_num = 0  # 上面参数种数
            left_num = 0  # 左边参数种数
            # 计算 left_num
            for i in table[0]:
                if not i:
                    left_num += 1
                else:
                    left_num += 1
                    break
            # 计算 up_num, up_paras
            for i in table:
                if '\\' not in i[left_num - 1]:
                    up_num += 1
                    up_paras.append(i[left_num - 1])
                else:
                    up_paras.append(i[left_num - 1].split('\\')[1])
                    up_num += 1
                    break
            # 计算 left_paras
            left_paras += table[up_num - 1][:left_num]
            left_paras[-1] = left_paras[-1].split('\\')[0]
            # 计算 up_paras_V
            next_num = len(table[0]) - left_num  # 下一个up参数的值数量
            for i in table[:up_num]:
                up_paras_V.append([])
                for j in i[left_num: left_num + next_num]:
                    if j:
                        next_num = 1
                        up_paras_V[-1].append(j)
                    else:
                        next_num += 1
            # 计算 left_paras_V
            next_num = len(table) - up_num  # 下一个left参数的值数量
            for i in range(left_num):
                left_paras_V.append([])
                for j in range(up_num, up_num + next_num):
                    k = table[j][i]
                    if k:
                        next_num = 1
                        left_paras_V[-1].append(k)
                    else:
                        next_num += 1
            # 构建全排列
            paras_v = [[left_paras[i], left_paras_V[i]] for i in range(left_num)]
            paras_v += [[up_paras[i], up_paras_V[i]] for i in range(up_num)]  # [[参数,(参数值,..)],..]
            error = False
            for i, j in paras_v:  # 查看表格参数是否错误
                if i not in self._paras_v_D or len(self._paras_v_D[i]) != len(j) or set(self._paras_v_D[i]) != set(j):
                    error = True
                    break
            if error or len(paras_v) != len(self._paras_v):
                out['error_table'] += 1
                continue
            all_p = self._all_permutation(paras_v)
            # 内容输入
            y = 0  # 第几个排列参数
            for i in range(up_num, len(table)):
                for j in range(left_num, len(table[0])):
                    x = {x[0]: all_p[y][i] for i, x in enumerate(paras_v)}
                    y += 1
                    try:
                        if not table[i][j] or table[i][j] == 'None':
                            continue
                    except:
                        raise ValueError('第%d个要读取的表中: table[%d][%d]不存在, 注意制表符和行数量!' % (ii + 1, i, j))
                    for k, v in self.setParaV_c(x, table[i][j], setNoCompute).items():
                        if k not in out:
                            out[k] = v
                        else:
                            out[k] += v
        if simplifyOut:
            out_ = {}
            for i, j in out.items():
                if len(str(j)) > 100:
                    continue
                else:
                    out_[i] = j
            out = out_
        return out

    def readParaLog(self, path, cv: list, spMark='=' * 50, paraMark='parameters: ', cMark='results: ',
                    setNoCompute=False, checkP=None, simplifyOut=True):
        r"""
        用于将日志中的内容读取到表格中, 日志例子:
        ...
        spMark\n
        paraMark{参数:参数值,..}\n
        cMark{内容名:内容值,..}\n
        ...
        @param path: 日志路径
        @param cv: str, 读取哪些内容名的值, 如果多个则顺序与ignoreP参数要求一致
        @param spMark: str, 每组参数的上分隔符, 第一个分隔符上面的不看
        @param paraMark: str, 参数字典的前缀标记, 必须顶行
        @param cMark: str, 内容字典的前缀标记, 必须顶行
        @param setNoCompute: bool, 日志中的值是否设置为非计算, 即nextParas是否输出
        @param checkP: {参数:参数值,..}, 用于检查日志中的其他参数是否一致
        @param simplifyOut: bool, 用于简化返回参数, 如果为True则不返回字符数超过100的统计参数
        @return: dict, 统计, 如果有的话success+repeat=非计算数量
        """
        if checkP is None:
            checkP = {}
        out = {
            'success': 0,  # 成功不重复写入的参数组数量
            'repeat': 0,  # 和已有值重复的写入数量
            'error': 0,  # 无法写入的参数组数量
            'error_seq': [],  # 无法写入的参数组
            'invalid': 0,  # 没有参数或内容的日志项数量
            'invalid_index': [],  # 没有参数或内容的日志项的索引, 从1开始
            'checkp_no_eq': 0,  # 检查额外参数不相等的参数组数量
        }
        checkP = self._allToString(checkP)
        try:
            with open(path, 'r', encoding='utf-8') as r:
                logs = r.read().split(spMark)[1:]
        except:
            return out
        for k, log in enumerate(logs):
            log = log.strip().split('\n')
            parasV, contentsV = {}, []
            for line in log:
                # line = line.strip()  # 不允许空格开头
                if line[:len(paraMark)] == paraMark:
                    parasV = ast.literal_eval(line.split(paraMark)[1])
                if line[:len(cMark)] == cMark:
                    contentsV = ast.literal_eval(line.split(cMark)[1])
                    contentsV = [contentsV[i] for i in cv]
            if parasV and contentsV:
                parasV = self._allToString(parasV)
                contentsV = self._allToString(contentsV)
                eq = True
                for i, j in checkP.items():
                    if i not in parasV or parasV[i] != j:
                        eq = False
                        break
                if eq:  # check 参数相等
                    for i, j in self.setParaV_c(parasV, contentsV, setNoCompute).items():
                        if i not in out:
                            out[i] = j
                        else:
                            out[i] += j
                else:  # 其他参数不相等
                    out['checkp_no_eq'] += 1
            else:
                out['invalid'] += 1
                out['invalid_index'].append(k + 1)
        if simplifyOut:
            out_ = {}
            for i, j in out.items():
                if len(str(j)) > 100:
                    continue
                else:
                    out_[i] = j
            out = out_
        return out


def paraTex_test():
    path = 'test/aa_result_test.log'
    # paraTex = ParaTex(
    #     leftParas=[['a', 'A', [1, 2], [10, 20]], ['b', 'B', [3, 8], [30, 80]]],
    #     upParas=[['c', 'C', [4], [40]], ['d', 'D', [5, 6, 7], [50, 60, 70]]],
    #     # blacklist=[{'a': (1,), 'c': (4,)}, {}],
    #     # whitelist=[{'a': (2,), 'c': (4,)}, {}],
    #     ignoreP=['b', 'd'],
    # )
    # print('paraTex._paras_v:', paraTex._paras_v)
    # print('paraTex._all_p:', paraTex._all_p)
    # print('paraTex._all_p_v8m:', paraTex._all_p_v8m)
    # for x in paraTex.nextParas():
    #     print(x)
    #     paraTex.setParaV_c(x, [1, 2, 3, 4, 5, 6])
    # print('paraTex._all_p_v8m:', paraTex._all_p_v8m)

    # paraTex = ParaTex(
    #     leftParas=[['a', 'A', [1, 2], [10, 20]], ['b', 'B', [3, 4], [30, 40]], ['f', 'F', [12, 13], [120, 130]]],
    #     upParas=[['c', 'C', [8, 9], [80, 90]], ['e', 'E', [10, 11], [100, 110]],
    #              ['d', 'D', [5, 6, 7], [50, 60, 70]]],
    #     # blacklist=[{'a': (1,), 'c': (4,)}, {}],
    #     # whitelist=[{'a': (2,), 'c': (4,)}, {}],
    # )
    # paraTex = ParaTex(
    #     leftParas=[['a', 'A', [1, 2], [10, 20]]],
    #     upParas=[['c', 'C', [4, 8, 9], [40, 80, 90]], ['b', 'B', [3, 4], [30, 40]]],
    #     # blacklist=[{'a': (1,), 'c': (4,)}, {}],
    #     # whitelist=[{'a': (2,), 'c': (4,)}, {}],
    # )
    # paraTex = ParaTex(
    #     leftParas=[['a', 'A', [1, 2], ['a1', 'a2']], ['b', 'B', [1, 2, 3], ['b1', 'b2', 'b3']],
    #                ['c', 'C', [1, 2, 3, 4, 5, 6], ['c1', 'c2', 'c3', 'c4', 'c5', 'c6']]],
    #     upParas=[['h', 'H', [1, 2], ['h1', 'h2']], ['i', 'I', [1, 2], ['i1', 'i2']],
    #              ['j', 'J', [1, 2], ['j1', 'j2']], ['k', 'K', [1, 2], ['k1', 'k2']]],
    # )
    # i = 0.001
    # n = 0
    # for x in paraTex.nextParas():
    #     x.update({12: 12})
    #     paraTex.setParaV_c(x, str(i))
    #     if n == 10:
    #         paraTex.setParaV_c(x, 3.1234)
    #     i += 0.001
    #     n += 1
    # # print('paraTex._leftParas:', paraTex._leftParas)
    # # print('paraTex._upParas:', paraTex._upParas)
    # print(paraTex.getTable(left_bf=False, up_bf=True, dec=4, if_bf=max, useV=False, path=path))
    # print(paraTex.getTableE(dec=4, useV=False, path=path))
    # print(paraTex.getParasG_n())
    paraTex = ParaTex(
        leftParas=[['h', 'H', [1, 2], ['h1', 'h2']], ['a', 'A', [1, 2], ['a1', 'a2']],
                   ['b', 'B', [1, 2, 3], ['b1', 'b2', 'b3']],
                   ['c', 'C', [1, 2, 3, 4, 5, 6], ['c1', 'c2', 'c3', 'c4', 'c5', 'c6']]],
        upParas=[['i', 'I', [1, 2], ['i1', 'i2']],
                 ['j', 'J', [1, 2], ['j1', 'j2']], ['k', 'K', [1, 2], ['k1', 'k2']]],
    )
    print(paraTex.readTableE(path, setNoCompute=True))
    print(paraTex.getTableE(dec=4, useV=False))
    print(paraTex.getParasG_n())


if __name__ == '__main__':
    paraTex_test()



from openpyxl.utils import column_index_from_string, get_column_letter
from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.cell.cell import Cell
from openpyxl.worksheet.merge import MergedCellRange
from typing import List, Generator, Tuple, Union, Optional, Dict, Hashable, TypedDict
from tsc_base import pair_to_dict, dict_to_pair
from pprint import pprint

NestedDict = Dict[Hashable, Union[int, 'NestedDict']]


class XlsxReader:
    @staticmethod
    def find_merged_cell(sheet: Worksheet, cell: Cell) -> Union[MergedCellRange, None]:
        """查找单元格是否是合并单元格的一部分

        Args:
            sheet (Worksheet): 工作表
            cell (Cell): 单元格

        Returns:
            Union[MergedCellRange, None]: 如果是合并单元格的一部分，返回合并单元格的范围，否则返回 None
        """
        for merged_range in sheet.merged_cells.ranges:
            if cell.coordinate in merged_range:
                return merged_range
        return None

    @classmethod
    def get_table_header(
        cls,
        min_col: Union[int, str],
        min_row: int, 
        max_col: Union[int, str],
        max_row: int, 
        sheet: Worksheet, 
        horizontal: bool = True, 
        root_list: List[list] = None,
        cell_len: int = 1,
    ) -> Generator[Tuple[list, int], None, None]:
        """使用 openpyxl 从 excel 的 sheet 中获取层次结构的表头

        Args:
            min_col (Union[int, str]): 范围，最小列，从1或A开始
            min_row (int): 范围，最小行，从1开始
            max_col (Union[int, str]): 范围，最大列，从1或A开始
            max_row (int): 范围，最大行，从1开始
            sheet (Worksheet): 工作表
            horizontal (bool, optional): 是否认为表头是从上到下的，否则是从左到右的
            root_list (List[list], optional): 递归时的根列表，无需传入
            cell_len (int, optional): 递归时的单元格长度/高度，无需传入

        Yields:
            Generator[Tuple[list, int], None, None]: 生成器，每次返回一个表头列表和单元格长度/高度
        """
        if isinstance(min_col, str):
            min_col = column_index_from_string(min_col)
        if isinstance(max_col, str):
            max_col = column_index_from_string(max_col)
        root_list = [] if root_list is None else root_list
        while min_col <= max_col and min_row <= max_row:
            cell: Cell = sheet.cell(row=min_row, column=min_col)
            merged_range = cls.find_merged_cell(sheet, cell)
            share_args = (sheet, horizontal, root_list + [cell.value])
            if merged_range:
                assert cell.column == merged_range.min_col, f"{cell.coordinate}, {merged_range.bounds}"
                assert min_row == merged_range.min_row, f"{cell.coordinate}, {merged_range.bounds}"
                if horizontal:
                    bounds = (min_col, merged_range.max_row + 1, merged_range.max_col, max_row)
                    yield from cls.get_table_header(*bounds, *share_args, merged_range.max_col - merged_range.min_col + 1)
                    min_col = merged_range.max_col + 1
                else:
                    bounds = (merged_range.max_col + 1, min_row, max_col, merged_range.max_row)
                    yield from cls.get_table_header(*bounds, *share_args, merged_range.max_row - merged_range.min_row + 1)
                    min_row = merged_range.max_row + 1
            else:
                if horizontal:
                    bounds = (min_col, min_row+1, min_col, max_row)
                    yield from cls.get_table_header(*bounds, *share_args, 1)
                    min_col += 1
                else:
                    bounds = (min_col+1, min_row, max_col, min_row)
                    yield from cls.get_table_header(*bounds, *share_args, 1)
                    min_row += 1
        if horizontal and min_row > max_row:
            yield root_list, cell_len
        elif not horizontal and min_col > max_col:
            yield root_list, cell_len

    @classmethod
    def get_docs_from_sheet(
        cls,
        sheet: Worksheet,
        th_end: Union[int, str],
        side_th_end: Union[int, str] = 0,
        horizontal: bool = True,
        start_col: Optional[Union[int, str]] = None,
        start_row: Optional[int] = None,
        end_col: Optional[Union[int, str]] = None,
        end_row: Optional[int] = None,
    ) -> dict:
        """从工作表中获取表格内容

        Args:
            sheet (Worksheet): 工作表
            th_end (Union[int, str]): 表头结束的位置，从1或A开始，horizontal为True时是数字行号，否则是字母列号
                比 start_col(horizontal=False) 或 start_row(horizontal=True) 小表示没有表头
            side_th_end (Union[int, str], optional): 侧边表头结束的位置，从1或A开始，horizontal为True时是字母列号，否则是数字行号
                比 start_col(horizontal=False) 或 start_row(horizontal=True) 小表示没有侧边表头
            horizontal (bool, optional): 表头是否是从上到下的，否则是从左到右的
            start_col (Optional[Union[int, str]], optional): 表格开始的列位置，从A开始，1也行，为None表示自动计算
            start_row (Optional[int], optional): 表格开始的行位置，从1开始，为None表示自动计算
            end_col (Optional[Union[int, str]], optional): 表格结束的列位置，从A开始，1也行，为None表示自动计算
            end_row (Optional[int], optional): 表格结束的行位置，从1开始，为None表示自动计算

        Returns:
            dict: 返回包括表格内容和侧边表头，表格内容是一个列表，每个元素是一条数据，侧边表头是一个嵌套字典
        """
        start_col = start_col or sheet.min_column
        start_row = start_row or sheet.min_row
        end_col = end_col or sheet.max_column
        end_row = end_row or sheet.max_row
        if isinstance(start_col, str):
            start_col = column_index_from_string(start_col)
        if isinstance(end_col, str):
            end_col = column_index_from_string(end_col)
        if isinstance(th_end, str):
            th_end = column_index_from_string(th_end)
        if isinstance(side_th_end, str):
            side_th_end = column_index_from_string(side_th_end)
        # 明确end下限
        if horizontal:
            th_end = max(th_end or 0, start_row - 1)
            has_th = th_end >= start_row
            side_th_end = max(side_th_end or 0, start_col - 1)
            has_side_th = side_th_end >= start_col
        else:
            th_end = max(th_end or 0, start_col - 1)
            has_th = th_end >= start_col
            side_th_end = max(side_th_end or 0, start_row - 1)
            has_side_th = side_th_end >= start_row
        # 获取表头
        if has_th:
            if horizontal:
                min_col, min_row, max_col, max_row = side_th_end + 1, start_row, end_col, th_end
            else:
                min_col, min_row, max_col, max_row = start_col, side_th_end + 1, th_end, end_row
            th_keys_len_L = list(cls.get_table_header(min_col, min_row, max_col, max_row, sheet, horizontal))
            th_len_D: NestedDict = pair_to_dict(th_keys_len_L)
            assert len(th_keys_len_L) == sum(1 for _ in dict_to_pair(th_len_D)), f"表头同一级可能存在同名现象: {th_len_D}"
        else:
            th_keys_len_L: List[Tuple[list, int]] = []
            th_len_D = {}
        # 获取侧边表头
        if has_side_th:
            if horizontal:
                min_col, min_row, max_col, max_row = start_col, th_end + 1, side_th_end, end_row
            else:
                min_col, min_row, max_col, max_row = th_end + 1, start_row, end_col, side_th_end
            side_th_keys_len_L = list(cls.get_table_header(min_col, min_row, max_col, max_row, sheet, not horizontal))
            side_th: NestedDict = pair_to_dict(side_th_keys_len_L)
            assert len(side_th_keys_len_L) == sum(1 for _ in dict_to_pair(side_th)), f"侧边表头同一级可能存在同名现象: {side_th}"
        else:
            side_th = {}
        # 获取表格内容
        td_L_L: List[list] = []  # 与 horizontal 无关的结果，1个list是一条数据
        if horizontal:
            for r in range(th_end + 1, end_row + 1):
                td_L = []
                for c in range(side_th_end + 1, end_col + 1):
                    td_L.append(sheet.cell(row=r, column=c).value)
                td_L_L.append(td_L)
        else:
            for c in range(th_end + 1, end_col + 1):
                td_L = []
                for r in range(side_th_end + 1, end_row + 1):
                    td_L.append(sheet.cell(row=r, column=c).value)
                td_L_L.append(td_L)
        # 生成 doc_L
        doc_L: List[dict] = []
        if has_th:
            for td_L in td_L_L:
                assert len(td_L) == len(th_keys_len_L), f"表格内容与表头长度不一致: {td_L}, {th_keys_len_L}"
                th_keys_v_L = []
                for v, (keys, l) in zip(td_L, th_keys_len_L):
                    assert l == 1, f"主表头最后一行不应该有合并单元格: {keys}, {l}"
                    th_keys_v_L.append((keys, v))
                doc_L.append(pair_to_dict(th_keys_v_L))
        return {
            'docs': doc_L,
            'side_th': side_th,
            'td_list': td_L_L,
            'th': th_len_D,
            'th_pairs': th_keys_len_L,
        }


if __name__ == "__main__":
    wb = load_workbook('test/openpyxl.xlsx')
    for name, th_end, side_th_end, horizontal in [
        ('横向', 3, 0, True),
        ('纵向', 'c', 0, False),
        ('横向2', 3, 3, True),
        ('纵向2', 'c', 3, False),
    ]:
        print('=' * 5, name, th_end, side_th_end, horizontal)
        sheet = wb[name]
        ret = XlsxReader.get_docs_from_sheet(sheet, th_end, side_th_end, horizontal=horizontal)
        pprint(ret)
        print()



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import os

if os.path.exists('readme.md'):
    long_description = open('readme.md', 'r', encoding='utf8').read()
else:
    long_description = '教程: https://github.com/aitsc/tsc-excel'

setup(
    name='tsc-excel',
    version='0.3',
    description="excel 文件读写，以json形式",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='aitsc',
    license='GPLv3',
    url='https://github.com/aitsc/tsc-excel',
    keywords='tools',
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries',
    ],
    python_requires='>=3.7',
    install_requires=[
        'tsc-base',
        'xlwt',
        'openpyxl',
        'tqdm',
    ],
)
